"""Simple API Interface for Maggma."""

import warnings

warnings.warn(
    "The maggma.api module has been deprecated and will no longer be supported. "
    "Users are advised to migrate their maggma.api applications to emmet-api: "
    "https://github.com/materialsproject/emmet/tree/main/emmet-api",
    category=DeprecationWarning,
    stacklevel=2,
)
